package require Thread

namespace eval curl {
    namespace export download
    namespace export read

    variable tmpdir "/tmp/$config::UID"
    file mkdir $tmpdir
    variable download_dir "$env(HOME)/Downloads"
    variable executor [tpool::create -minworkers 8 -maxworkers 12]

    variable user_agent \
        "Mozilla/5.0 (X11; Linux x86_64; rv:81.0) Gecko/20100101 Firefox/81.0"

    proc threaded_download {url target} {
        variable user_agent
        variable executor

        set curl_cmd "curl -s -A {$user_agent} -f -L {$url} -o {$target}"
        set cmd "exec -ignorestderr -- $curl_cmd"

        set task_id [tpool::post -nowait $executor $cmd]
        return $task_id
    }

    proc wait_threaded_return {task_id} {
        variable executor

        set ready_count [llength [tpool::wait $executor $task_id]]
        if {$ready_count == 0} {return "WAIT"}

        try {
            set results [tpool::get $executor $task_id]
        } trap CHILDSTATUS {results options} {
            set status [lindex [dict get $options -errorcode] 2]
            puts "task $task_id status: $status"
            return "ERROR"
        }
        return "OK"
    }

    proc _download {url target} {
        variable user_agent

        set cmd "exec -ignorestderr -- curl -s -f -A {$user_agent} -L {$url} -o {$target}"

        try {
            set results [eval $cmd]
        } trap CHILDSTATUS {results options} {
            set status [lindex [dict get $options -errorcode] 2]
            puts "curl $url -o $target status: $status"
            return ""
        }
        return $target
    }

    proc download {url target {cached false}} {
        if {$cached == true} {
            set exists [file exists $target]
            if {$exists} {
                return $target
            }
        }
        return [_download $url $target]
    }

    proc _read {url {headers {}}} {
        variable user_agent

        set curl_command [list curl -A $user_agent -L --silent $url]

        foreach {k v} $headers {
            set curl_command [lappend curl_command -H "$k:$v"]
        }

        try {
            set results [eval "exec -ignorestderr -- $curl_command"]
        } trap CHILDSTATUS {results options} {
            set status [lindex [dict get $options -errorcode] 2]
            puts "curl $url status: $status"
            return ""
        }
        return $results
    }

    proc read {url {target ""} {headers {}}} {
        variable tmpdir

        if {$target != ""} {
            set target_path "$tmpdir/$target"
            set exists [file exists $target_path]
            if {$exists} {
                set fd [open $target_path r]
                set content [::read $fd]
                close $fd
                return $content
            }
        }

        set results [curl::_read $url $headers]
        if {$target != ""} {
            set fd [open $target_path w]
            puts -nonewline $fd $results
            close $fd
        }
        return $results
    }

    # ------------------------------
    # With HTTP status codes:
    #
    proc _http_read {url {headers {}}} {
        variable user_agent

        set curl_command [list curl -i -A $user_agent -L --silent $url]

        foreach {k v} $headers {
            set curl_command [lappend curl_command -H "$k:$v"]
        }

        try {
            set response [eval "exec -ignorestderr -- $curl_command"]
        } trap CHILDSTATUS {response options} {
            set status [lindex [dict get $options -errorcode] 2]
            return ""
        }
        return $response
    }

    proc parse_http_response {text} {
        set reading_headers true
        set status 0
        set headers [dict create]
        set body ""
        set counter 0
        foreach {line} [split $text "\n"] {
            incr counter
            set line [string trimright $line "\r"]

            if {$counter == 1} {
                # HTTP/2 200
                # HTTP/1.1 200 Ok
                set parts [split $line " "]
                set status [lindex $parts 1]
                continue
            }

            if {$reading_headers == true} {
                if {$line == ""} {
                    if {$status == "301"} {
                        set status 0
                        set counter 0
                        set headers [dict create]
                    } else {
                        set reading_headers false
                        continue
                    }
                } else {
                    set f [string first ": " $line]
                    set key [string range $line 0 $f-1]
                    set value [string range $line $f+2 end]
                    dict set headers $key $value
                }
            } else {
                set body "$body$line"
            }
        }

        return [dict create \
            "status" $status \
            "headers" $headers \
            "body" $body]
    }
    proc http_read {url {target ""} {headers {}}} {
        variable tmpdir

        if {$target != ""} {
            set target_path "$tmpdir/$target"
            set exists [file exists $target_path]
            if {$exists} {
                set fd [open $target_path r]
                set content [::read $fd]
                close $fd
                return [parse_http_response $content]
            }
        }

        set response [curl::_http_read $url $headers]
        if {$target != ""} {
            set fd [open $target_path w]
            puts -nonewline $fd $response
            close $fd
        }
        return [parse_http_response $response]
    }
    proc http_send {method url headers body} {
        set curl_cmd "curl -X $method -s -A {$user_agent} -f -L {$url} -H {Content-Type: application/json} -d {$body}"
        set cmd "exec -ignorestderr -- $curl_cmd"
    }
}
